from __future__ import annotations
from typing import Any, Iterable, Mapping, Optional
from frame_semantic_transformer.data.framenet import get_fulltext_docs

from frame_semantic_transformer.data.sesame_data_splits import (
    SESAME_DEV_FILES,
    SESAME_TEST_FILES,
)
from frame_semantic_transformer.data.tasks import (
    ArgumentsExtractionSample,
    ArgumentsExtractionTask,
    FrameClassificationSample,
    FrameClassificationTask,
    TaskSample,
    TriggerIdentificationSample,
    TriggerIdentificationTask,
)


def load_framenet_samples(
    include_docs: Optional[Iterable[str]] = None,
    exclude_docs: Optional[Iterable[str]] = None,
) -> list[TaskSample]:
    samples: list[TaskSample] = []
    for doc in get_fulltext_docs():
        if exclude_docs and doc["filename"] in exclude_docs:
            continue
        if include_docs and doc["filename"] not in include_docs:
            continue
        samples += parse_samples_from_fulltext_doc(doc)
    return samples


def load_sesame_train_samples() -> list[TaskSample]:
    return load_framenet_samples(exclude_docs=SESAME_DEV_FILES + SESAME_TEST_FILES)


def load_sesame_test_samples() -> list[TaskSample]:
    return load_framenet_samples(include_docs=SESAME_TEST_FILES)


def load_sesame_dev_samples() -> list[TaskSample]:
    return load_framenet_samples(include_docs=SESAME_DEV_FILES)


def parse_frame_samples_from_annotation_set(
    annotation_set: Iterable[Mapping[str, Any]]
) -> list[FrameClassificationSample | ArgumentsExtractionSample]:
    """
    Helper to parse sample sentences out of framenet annotation sets
    Not all annotation sets contain frames, so this will filter out any invalid annotation sets
    ex: lu = fn.lus()[0]; samples = parse_samples_from_exemplars(lu.exemplars)
    """
    sample_sentences: list[FrameClassificationSample | ArgumentsExtractionSample] = []
    for annotation in annotation_set:
        if "FE" in annotation and "Target" in annotation and "frame" in annotation:
            for trigger_loc in annotation["Target"]:
                sample_sentences.append(
                    FrameClassificationSample(
                        task=FrameClassificationTask(
                            text=annotation["text"],
                            trigger_loc=trigger_loc[0],
                        ),
                        frame=annotation["frame"]["name"],
                    )
                )
                sample_sentences.append(
                    ArgumentsExtractionSample(
                        task=ArgumentsExtractionTask(
                            text=annotation["text"],
                            trigger_loc=trigger_loc[0],
                            frame=annotation["frame"]["name"],
                        ),
                        frame_element_locs=annotation["FE"][0],
                    )
                )

    return sample_sentences


def parse_trigger_sample_from_annotation_set(
    annotation_set: Iterable[Mapping[str, Any]]
) -> TriggerIdentificationSample | None:
    """
    Helper to parse sample sentences out of framenet annotation sets
    Not all annotation sets contain frames, so this will filter out any invalid annotation sets
    ex: lu = fn.lus()[0]; samples = parse_samples_from_exemplars(lu.exemplars)
    """
    target_locs = []
    text: Optional[str] = None
    for annotation in annotation_set:
        if "Target" in annotation:
            text = annotation["text"]
            for loc in annotation["Target"]:
                target_locs.append(loc[0])

    if not text or len(target_locs) == 0:
        return None
    return TriggerIdentificationSample(
        task=TriggerIdentificationTask(text=text), trigger_locs=target_locs
    )


def parse_samples_from_fulltext_doc(doc: Mapping[str, Any]) -> list[TaskSample]:
    """
    Helper to parse sample sentences out of framenet exemplars, contained in lexical units
    ex: lu = fn.lus()[0]; samples = parse_samples_from_exemplars(lu)
    """
    samples: list[TaskSample] = []
    for sentence in doc["sentence"]:
        samples += parse_frame_samples_from_annotation_set(sentence["annotationSet"])
        trigger_sample = parse_trigger_sample_from_annotation_set(
            sentence["annotationSet"]
        )
        if trigger_sample:
            samples.append(trigger_sample)
    return samples



from __future__ import annotations
from functools import lru_cache
from typing import Any, Sequence, Mapping
import nltk

from nltk.corpus import framenet as fn


class InvalidFrameError(Exception):
    pass


def ensure_framenet_downloaded() -> None:
    try:
        nltk.data.find("corpora/framenet_v17")
    except LookupError:
        nltk.download("framenet_v17")


def is_valid_frame(frame: str) -> bool:
    return frame in get_all_valid_frame_names()


def get_core_frame_elements(frame: str) -> list[str]:
    if not is_valid_frame(frame):
        raise InvalidFrameError(frame)
    return get_frame_elements_map_by_core_type()[frame]["core"]


def get_non_core_frame_elements(frame: str) -> list[str]:
    if not is_valid_frame(frame):
        raise InvalidFrameError(frame)
    return get_frame_elements_map_by_core_type()[frame]["noncore"]


@lru_cache(1)
def get_frame_elements_map_by_core_type() -> dict[str, dict[str, list[str]]]:
    """
    fast-lookup helper for frames -> frame elements for faster eval
    """
    results: dict[str, dict[str, list[str]]] = {}
    for frame in fn.frames():
        results[frame.name] = {"core": [], "noncore": []}
        for element_name, element in frame.FE.items():
            element_type = "core" if element.coreType == "Core" else "noncore"
            results[frame.name][element_type].append(element_name)
    return results


@lru_cache(1)
def get_all_valid_frame_names() -> set[str]:
    return {frame.name for frame in fn.frames()}


def get_fulltext_docs() -> Sequence[Mapping[str, Any]]:
    return fn.docs()



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Sequence

from frame_semantic_transformer.data.tasks.Task import Task


class TaskSample(ABC):
    """
    Abstract interface for all Task Samples
    """

    task: Task

    def get_task_name(self) -> str:
        return self.task.get_task_name()

    def get_input(self) -> str:
        return self.task.get_input()

    @abstractmethod
    def get_target(self) -> str:
        pass

    @staticmethod
    @abstractmethod
    def evaluate_prediction(
        prediction_outputs: Sequence[str], target: str, input: str
    ) -> tuple[float, float, float]:
        "return a tuple indicating the number of true positives, false positives, and false negatives in the prediction"
        pass



from __future__ import annotations
import re
from typing import Iterator, Sequence, TypeVar
from torch import Tensor

from transformers import T5Tokenizer

from frame_semantic_transformer.constants import PADDING_LABEL_ID

T = TypeVar("T")


def chunk_list(lst: Sequence[T], chunk_size: int) -> Iterator[Sequence[T]]:
    for i in range(0, len(lst), chunk_size):
        yield lst[i : i + chunk_size]


def standardize_punct(sent: str) -> str:
    """
    Try to standardize things like "He 's a man" -> "He's a man"
    """
    updated_sent = T5Tokenizer.clean_up_tokenization(sent)
    # remove space before punct
    updated_sent = re.sub(r"([a-zA-Z0-9])\s+(\*?[.',:?])", r"\1\2", updated_sent)
    # remove repeated *'s
    updated_sent = re.sub(r"\*+", "*", updated_sent)
    # fix spaces in contractions
    updated_sent = re.sub(r"([a-zA-Z0-9])\s+(\*?n't)", r"\1\2", updated_sent)
    # remove ``
    updated_sent = re.sub(r"\s*``\s*", " ", updated_sent)
    # replace "*n't" with "n*'t, for the tokenizer"
    updated_sent = re.sub(r"\*n't", "n*'t", updated_sent)
    # put a space between * and letter chars, since it seems to work better with the tokenizer
    updated_sent = re.sub(r"\*([a-zA-Z0-9])", r"* \1", updated_sent)

    return updated_sent.strip()


def marked_string_to_locs(
    text: str, symbol: str = "*", remove_spaces: bool = True
) -> tuple[str, list[int]]:
    """
    Take a string like "He * went to the * store" and return the indices of the tagged words,
    in this case "went" and "store", and remove the tags (in this case the *'s)
    """
    output_str = ""
    remaining_str = text
    locs: list[int] = []
    symbol_index = remaining_str.find("*")

    while symbol_index != -1:
        locs.append(symbol_index + len(output_str))
        output_str += remaining_str[:symbol_index]
        remaining_str = remaining_str[symbol_index + len(symbol) :]
        if remove_spaces:
            remaining_str = remaining_str.strip()
        symbol_index = remaining_str.find("*")
    output_str += remaining_str
    return output_str, locs


def trim_batch(
    input_ids: Tensor, attention_mask: Tensor, labels: Tensor
) -> tuple[Tensor, Tensor, Tensor]:
    """
    Helper to trim a batch of inputs / labels to strip padding down to the length of the longest item in the batch
    This helps the model run faster by avoiding needing to generate up to 512 characters of padding when the meaningful
    content is much shorter.
    """
    longest_inputs = int(attention_mask.sum(dim=1).max().item())
    longest_labels = int((labels != PADDING_LABEL_ID).sum(dim=1).max().item())

    truncated_input_ids = input_ids[:, :longest_inputs].contiguous()
    truncated_attention_mask = attention_mask[:, :longest_inputs].contiguous()
    truncated_labels = labels[:, :longest_labels].contiguous()

    return (truncated_input_ids, truncated_attention_mask, truncated_labels)



from __future__ import annotations
from dataclasses import dataclass
from typing import Sequence

from .FrameClassificationTask import FrameClassificationTask
from .TaskSample import TaskSample


@dataclass
class FrameClassificationSample(TaskSample):
    task: FrameClassificationTask
    frame: str

    # -- input / target for training --

    def get_target(self) -> str:
        return self.frame

    @staticmethod
    def evaluate_prediction(
        prediction_outputs: Sequence[str], target: str, _input: str
    ) -> tuple[float, float, float]:
        prediction = FrameClassificationTask.parse_output(prediction_outputs)
        if prediction and prediction == target:
            return (1.0, 0.0, 0.0)
        else:
            # sesame treats any non-correct frame as both a false pos and false neg
            # https://github.com/swabhs/open-sesame/blob/master/sesame/evaluation.py#L67
            return (0.0, 1.0, 1.0)



from .chain_augmentations import chain_augmentations
from .DataAugmentation import DataAugmentation
from .LowercaseAugmentation import LowercaseAugmentation
from .RemoveContractionsAugmentation import RemoveContractionsAugmentation
from .RemoveEndPunctuationAugmentation import RemoveEndPunctuationAugmentation

__all__ = (
    "chain_augmentations",
    "DataAugmentation",
    "LowercaseAugmentation",
    "RemoveContractionsAugmentation",
    "RemoveEndPunctuationAugmentation",
)
